# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""Subpackage containing container classes for multidimensional fields and ways to create them."""


from .field import *
from .shapes import *
from .vectors import *


__all__ = []
__all__.extend(field.__all__)
__all__.extend(shapes.__all__)
__all__.extend(vectors.__all__)


del field
del shapes
del vectors



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""Subpackage containing utility functionality."""


from .quaternion import *


__all__ = []
__all__.extend(quaternion.__all__)


del quaternion



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides the :class:`~.Quaternion` class which can be used for rotations."""


import logging

import numpy as np


__all__ = ['Quaternion']


class Quaternion(object):
    R"""Class representing a rotation expressed by a quaternion.

    A quaternion is a four-dimensional description of a rotation which can also be described by
    a rotation vector (`v1`, `v2`, `v3`) and a rotation angle :math:`\theta`. The four components
    are calculated to:

    .. math::

        w = \cos(\theta/2)
        x = v_1 \cdot \sin(\theta/2)
        y = v_2 \cdot \sin(\theta/2)
        z = v_3 \cdot \sin(\theta/2)

    Use the :func:`~.from_axisangle` and :func:`~.to_axisangle` to convert to axis-angle
    representation and vice versa. Quaternions can be multiplied by other quaternions, which
    results in a new rotation or with a vector, which results in a rotated vector.

    Attributes
    ----------
    values : float
        The four quaternion values `w`, `x`, `y`, `z`.

    """

    NORM_TOLERANCE = 1E-6

    _log = logging.getLogger(__name__ + '.Quaternion')

    @property
    def conj(self):
        """The conjugate of the quaternion, representing a tilt in opposite direction."""
        w, x, y, z = self.values
        return Quaternion((w, -x, -y, -z))

    @property
    def matrix(self):
        """The rotation matrix representation of the quaternion."""
        w, x, y, z = self.values
        return np.array([[1 - 2 * (y ** 2 + z ** 2), 2 * (x * y - w * z), 2 * (x * z + w * y)],
                         [2 * (x * y + w * z), 1 - 2 * (x ** 2 + z ** 2), 2 * (y * z - w * x)],
                         [2 * (x * z - w * y), 2 * (y * z + w * x), 1 - 2 * (x ** 2 + y ** 2)]])

    def __init__(self, values):
        self._log.debug('Calling __init__')
        self.values = values
        self._normalize()
        self._log.debug('Created ' + str(self))

    def __mul__(self, other):  # self * other
        self._log.debug('Calling __mul__')
        if isinstance(other, Quaternion):  # Quaternion multiplication
            return self.dot_quat(self, other)
        elif len(other) == 3:  # vector multiplication (Caution: normalises!)
            q_vec = Quaternion((0,) + tuple(other))
            q = self.dot_quat(self.dot_quat(self, q_vec), self.conj)
            return q.values[1:]

    def _normalize(self):
        self._log.debug('Calling _normalize')
        mag2 = np.sum(n ** 2 for n in self.values)
        if abs(mag2 - 1.0) > self.NORM_TOLERANCE:
            mag = np.sqrt(mag2)
            self.values = tuple(n / mag for n in self.values)

    def dot_quat(self, q1, q2):
        """Multiply two :class:`~.Quaternion` objects to create a new one (always normalized).

        Parameters
        ----------
        q1, q2 : :class:`~.Quaternion`
            The quaternion which should be multiplied.

        Returns
        -------
        quaternion : :class:`~.Quaternion`
            The resulting quaternion.

        """
        self._log.debug('Calling dot_quat')
        w1, x1, y1, z1 = q1.values
        w2, x2, y2, z2 = q2.values
        w = w1 * w2 - x1 * x2 - y1 * y2 - z1 * z2
        x = w1 * x2 + x1 * w2 + y1 * z2 - z1 * y2
        y = w1 * y2 + y1 * w2 + z1 * x2 - x1 * z2
        z = w1 * z2 + z1 * w2 + x1 * y2 - y1 * x2
        return Quaternion((w, x, y, z))

    @classmethod
    def from_axisangle(cls, vector, theta):
        """Create a quaternion from an axis-angle representation

        Parameters
        ----------
        vector : :class:`~numpy.ndarray` (N=3)
            Vector around which the rotation is executed.
        theta : float
            Rotation angle.

        Returns
        -------
        quaternion : :class:`~.Quaternion`
            The resulting quaternion.

        """
        cls._log.debug('Calling from_axisangle')
        x, y, z = vector
        theta /= 2.
        w = np.cos(theta)
        x *= np.sin(theta)
        y *= np.sin(theta)
        z *= np.sin(theta)
        return cls((w, x, y, z))

    def to_axisangle(self):
        """Convert the quaternion to axis-angle-representation.

        Returns
        -------
        vector, theta : :class:`~numpy.ndarray` (N=3), float
            Vector around which the rotation is executed and rotation angle.

        """
        self._log.debug('Calling to_axisangle')
        w, x, y, z = self.values
        theta = 2.0 * np.arccos(w)
        return np.array((x, y, z)), theta



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""Subpackage containing functionality for visualisation of multidimensional fields."""


from . import fields
from . import io
from . import models
from . import reconstruct
from . import vis
from . import utils
from .version import version as __version__
from .version import git_revision as __git_revision__

import logging
_log = logging.getLogger(__name__)
_log.info(f'Imported EMPyRe V-{__version__} GIT-{__git_revision__}')
del logging


__all__ = ['fields', 'io', 'models', 'reconstruct', 'vis', 'utils']


del version



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides helper functions to the vis module."""


import os
import glob
import shutil
import logging
from numbers import Number
from contextlib import contextmanager

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from ..fields.field import Field


__all__ = ['new', 'savefig', 'calc_figsize', 'use_style', 'copy_mpl_stylesheets']
_log = logging.getLogger(__name__)


def new(nrows=1, ncols=1, mode='image', figsize=None, textwidth=None, width_scale=1, aspect=None, **kwargs):
    R"""Convenience function for the creation of a new subplot grid (wraps `~matplotlib.pyplot.subplots`).

    If you use the `textwidth` parameter, plot sizes are fitting into publications with LaTeX. Requires two stylesheets
    `empyre-image` and `empyre-plot` corresponding to its two `mode` settings. Those stylesheets use
    `constrained_layout=True` to achieve well behaving plots without much whitespace around. This function should work
    fine for a small number of images (e.g. 1, 2x2, etc.), for more fine grained control, the contexts can be used
    directly if they are installed corretly, or use width_scale to build the images separately (e.g. 2 adjacent with
    width=0.5). For images, it is assumed that most images are square (and therefore `aspect=1`).

    Parameters
    ----------
    nrows : int, optional
        Number of rows of the subplot grid, by default 1
    ncols : int, optional
        Number of columns of the subplot grid, by default 1
    mode : {'image', 'plot'}, optional
        Mode of the new subplot grid, by default 'image'. Both modes have dedicated matplotlib styles which are used
        and which are installed together with EMPyRe. The 'image' mode disables axis labels and ticks, mainly intended
        to be used with `~matplotlib.pyplot.imshow` with `~empyre.vis.decorators.scalebar`, while the 'plot'
        mode should be used for traditional plots like with `~matplotlib.pyplot.plot` or `~matplotlib.pyplot.scatter`.
    figsize : (float, float), optional
        Width and height of the figure in inches, defaults to rcParams["figure.figsize"], which depends on the chosen
        stylesheet. If set, this will overwrite all other following parameters.
    textwidth : float, optional
        The textwidth of your LaTeX document in points, which you can get py using "\the\textwidth". If this is not None
        (the default), this will be used to define the figure size if it is not set explicitely.
    width_scale : float, optional
        Only meaningful if `textwidth` is set. If it is, `width_scale` will be a scaling factor for the figure width.
        Example: if you set this to 0.5, your figure will span half of the textwidth. Default is 1.
    aspect : float, optional
        Aspect ratio of the figure height relative to the figure width. If None (default), the aspect is set to be 1
        for `mode=image` and to 'golden' for `mode=plot`, which adjusts the aspect to represent the golden ratio of
        0.6180... If `ncols!=nrows`, it often makes sense to use `aspect=nrows/ncols` here.

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        The constructed figure.
    axes : axes.Axes object or array of Axes objects.
        axes can be either a single Axes object or an array of Axes objects if more than one subplot was created.
        The dimensions of the resulting array can be controlled with the squeeze keyword argument.

    Notes
    -----
    additional kwargs are passed to `~matplotlib.pyplot.subplots`.

    """
    _log.debug('Calling new')
    assert mode in ('image', 'plot'), "mode has to be 'image', or 'plot'!"
    if figsize is None and textwidth is not None:  # Only then is all this necessary:
        if aspect is None:
            aspect = 'golden' if mode == 'plot' else 1  # Both image modes have 'same' as default'!
        elif isinstance(aspect, Field):
            dim_uv = [d for d in aspect.dim if d != 1]
            assert len(dim_uv) == 2, f"Couldn't find fields aspect ({len(dim_uv)} squeezed dimensions, has to be of 2)!"
            aspect = dim_uv[0]/dim_uv[1]  # height/width
        else:
            assert isinstance(aspect, Number), 'aspect has to be None, a number or a field instance squeezable to 2D!'
        figsize = calc_figsize(width_scale=width_scale, aspect=aspect, textwidth=textwidth)
    if mode == 'image':
        with use_style('empyre-image'):
            return plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)
    else:  # mode == 'plot':
        with use_style('empyre-plot'):
            return plt.subplots(nrows=nrows, ncols=ncols, figsize=figsize, **kwargs)


def savefig(fname, **kwargs):
    """Utility wrapper around :func:`~matplotlib.pyplot.savefig` to save the current figure.

    Parameters
    ----------
    fname : str or PathLike or file-like object
        Path to the file wherein the figure should be saved.

    Notes
    -----
    Uses the 'empyre-save' stylesheet (installed together with EMPyRe to control the saving behaviour. Any kwargs are
    passed to :func:`~matplotlib.pyplot.savefig`.

    """
    _log.debug('Calling savefig')
    with use_style('empyre-save'):
        plt.savefig(fname, **kwargs)


def calc_figsize(textwidth, width_scale=1, aspect=1):
    R"""Helper function to calculate the figure size from various parameters. Useful for publications via LaTeX.

    Parameters
    ----------
    textwidth : float
        The textwidth of your LaTeX document in points, which you can get py using "\the\textwidth".
    width_scale : float, optional
        Scaling factor for the figure width. Example: if you set this to 0.5, your figure will span half of the
        textwidth. Default is 1.
    aspect : float, optional
        Aspect ratio of the figure height relative to the figure width. If None (default), the aspect is set to be 1
        for `mode=image` and to 'golden' for `mode=plot`, which adjusts the aspect to represent the golden ratio of
        0.6180...

    Returns
    -------
    figsize: (float, float)
        The determined figure size

    Notes
    -----
    Based on snippet from Florian Winkler.

    """
    _log.debug('Calling calc_figsize')
    GOLDEN_RATIO = (1 + np.sqrt(5)) / 2   # Aesthetic ratio!
    INCHES_PER_POINT = 1.0 / 72.27  # Convert points to inch, LaTeX constant, apparently...
    textwidth_in = textwidth * INCHES_PER_POINT  # Width of the text in inches
    fig_width = textwidth_in * width_scale  # Width in inches
    if aspect == 'golden':
        fig_height = fig_width / GOLDEN_RATIO
    elif isinstance(aspect, Number):
        fig_height = textwidth_in * aspect
    else:
        raise ValueError(f"aspect has to be either a number, 'same' or 'golden'! Was {aspect}!")
    fig_size = [fig_width, fig_height]  # Both in inches
    return fig_size


@contextmanager
def use_style(stylename):
    """Context that uses a matplotlib stylesheet. Can fall back to local mpl stylesheets if necessary!

    Parameters
    ----------
    stylename : str
        A style specification.

    Yields
    -------
    context
        Context manager for using style settings temporarily.

    """
    try:  # Try to load the style directly (works if it is installed somewhere mpl looks for it):
        with plt.style.context(stylename) as context:
            yield context
    except OSError:  # Stylesheet not found, use local ones:
        mplstyle_path = os.path.join(os.path.dirname(__file__), 'mplstyles', f'{stylename}.mplstyle')
        with plt.style.context(mplstyle_path) as context:
            yield context


def copy_mpl_stylesheets():
    """Copy matplotlib styles to the users matplotlib config directory. Useful if you want to utilize them elsewhere.

    Notes
    -----
    You might need to restart your Python session for the stylesheets to be recognized/found!

    """
    # Find matplotlib styles:
    user_stylelib_path = os.path.join(mpl.get_configdir(), 'stylelib')
    vis_dir = os.path.dirname(__file__)
    style_files = glob.glob(os.path.join(vis_dir, 'mplstyles', '*.mplstyle'))
    # Copy them to the local matplotlib styles folder:
    if not os.path.exists(user_stylelib_path):
        os.makedirs(user_stylelib_path)
    for style_path in style_files:
        _, fname = os.path.split(style_path)
        dest = os.path.join(user_stylelib_path, fname)
        shutil.copy(style_path, dest)



#!python
# coding=utf-8
"""Setup for testing, building, distributing and installing the 'EMPyRe'-package"""


import os
import subprocess
from setuptools import setup, config


# Read version from setup.py:
version = config.read_configuration('setup.cfg')['metadata']['version']
# Get current git revision:
try:
    git_rev = subprocess.check_output(['git', 'rev-parse', 'HEAD']).strip().decode()
except Exception:
    git_rev = "???"
# Write both to version.py:
print(R'writing src\empyre\version.py')
with open(os.path.join(os.path.dirname(__file__), 'src', 'empyre', 'version.py'), 'w') as vfile:
    vfile.write('# -*- coding: utf-8 -*-\n' +
                '""""This file was automatically generated by `setup.py`"""\n' +
                f'version = "{version}"\n' +
                f'git_revision = "{git_rev}"\n')

# Run setup (reads metadata & options from setup.cfg):
print(R'running setup.py')
setup()



# -*- coding: utf-8 -*-
# Copyright 2020 by Forschungszentrum Juelich GmbH
# Author: J. Caron
#
"""This module provides the miscellaneous helper functions."""


import logging
import itertools
from time import time

import numpy as np
from tqdm import tqdm
from scipy.spatial import cKDTree, qhull
from scipy.interpolate import LinearNDInterpolator


__all__ = ['levi_civita', 'interp_to_regular_grid']
_log = logging.getLogger(__name__)


def levi_civita(i, j, k):
    _log.debug('Calling levi_civita')
    return (j-i) * (k-i) * (k-j) / (np.abs(j-i) * np.abs(k-i) * np.abs(k-j))


def interp_to_regular_grid(points, values, scale, scale_factor=1, step=1, convex=True):
    """Interpolate values on points to regular grid.

    Parameters
    ----------
    points : np.ndarray, (N, 3)
        Array of points, describing the location of the values that should be interpolated. Three columns x, y, z!
    values : np.ndarray, (N, c)
        Array of values that should be interpolated to the new grid. `c` is the number of components (`1` for scalar
        fields, `3` for normal 3D vector fields).
    scale : tuple of 3 ints
        Scale along each of the 3 spatial dimensions. Usually given in nm.
    scale_factor : float, optional
        Additional scaling factor that should be used if the original points are not described on a nm-scale. Use this
        to convert from nm of `scale` to the unit of the points. By default 1.
    step : int, optional
        If this is bigger than 1 (the default), only every `step` point is taken into account. Can speed up calculation,
        but you'll lose accuracy of your interpolation.
    convex : bool, optional
        By default True. If this is set to False, additional measures are taken to find holes in the point cloud.
        WARNING: this is an experimental feature that should be used with caution!

    Returns
    -------
    interpolation: np.ndarray
        Interpolated grid with shape `(zdim, ydim, xdim)` for scalar and `(zdim, ydim, xdim, ncomp)` for vector field.

    """
    _log.debug('Calling interpolate_to_regular_grid')
    z_uniq = np.unique(points[:, 2])
    _log.info(f'unique positions along z: {len(z_uniq)}')
    #  Translate scale in local units (not necessarily nm), taken care of with `scale_factor`:
    scale = tuple([s * scale_factor for s in scale])
    # Determine the size of the point cloud of irregular coordinates:
    x_min, x_max = points[:, 0].min(), points[:, 0].max()
    y_min, y_max = points[:, 1].min(), points[:, 1].max()
    z_min, z_max = points[:, 2].min(), points[:, 2].max()
    x_diff, y_diff, z_diff = np.ptp(points[:, 0]), np.ptp(points[:, 1]), np.ptp(points[:, 2])
    _log.info(f'x-range: {x_min:.2g} <-> {x_max:.2g} ({x_diff:.2g})')
    _log.info(f'y-range: {y_min:.2g} <-> {y_max:.2g} ({y_diff:.2g})')
    _log.info(f'z-range: {z_min:.2g} <-> {z_max:.2g} ({z_diff:.2g})')
    # Determine dimensions from given grid spacing a:
    dim = tuple(np.round(np.asarray((z_diff/scale, y_diff/scale, x_diff/scale), dtype=int)))
    x = x_min + scale * (np.arange(dim[2]) + 0.5)  # +0.5: shift to pixel center!
    y = y_min + scale * (np.arange(dim[1]) + 0.5)  # +0.5: shift to pixel center!
    z = z_min + scale * (np.arange(dim[0]) + 0.5)  # +0.5: shift to pixel center!
    # Create points for new Euclidian grid; fliplr for (x, y, z) order:
    points_euc = np.fliplr(np.asarray(list(itertools.product(z, y, x))))
    # Make values 2D (if not already); double .T so that a new axis is added at the END (n, 1):
    values = np.atleast_2d(values.T).T
    # Prepare interpolated grid:
    interpolation = np.empty(dim+(values.shape[-1],), dtype=np.float)
    _log.info(f'Dimensions of new grid: {interpolation.shape}')
    # Calculate the Delaunay triangulation (same for every component of multidim./vector fields):
    _log.info('Start Delaunay triangulation...')
    tick = time()
    triangulation = qhull.Delaunay(points[::step])
    tock = time()
    _log.info(f'Delaunay triangulation complete (took {tock-tick:.2f} s)!')
    # Perform the interpolation for each column of `values`:
    for i in tqdm(range(values.shape[-1])):
        # Create interpolator for the given triangulation and the values of the current column:
        interpolator = LinearNDInterpolator(triangulation, values[::step, i], fill_value=0)
        # Interpolate:
        interpolation[..., i] = interpolator(points_euc).reshape(dim)
    # If NOT convex, we have to check for additional holes in the structure (EXPERIMENTAL):
    if not convex:  # Only necessary if the user expects holes in the (-> nonconvex) distribution:
        # Create k-dimensional tree for queries:
        tree = cKDTree(points)
        # Query the tree for nearest neighbors, x: points to query, k: number of neighbors, p: norm
        # to use (here: 2 - Euclidean), distance_upper_bound: maximum distance that is searched!
        data, leafsize = tree.query(x=points_euc, k=1, p=2, distance_upper_bound=2*np.mean(scale))
        # Create boolean mask that determines which interpolation points have no neighbor near enough:
        mask = np.isinf(data).reshape(dim)  # Points further away than upper bound were marked 'inf'!
        for i in tqdm(range(values.shape[-1])):  # Set these points to zero (NOTE: This can take a looooong time):
            interpolation[..., i].ravel()[mask.ravel()] = 0
    return np.squeeze(interpolation)
